"""Definition of an ElkM1 Zone"""
from .const import (
    Max,
    TextDescriptions,
    ZoneType,
    ZoneLogicalStatus,
    ZonePhysicalStatus,
)
from .elements import Element, Elements
from .message import az_encode, zd_encode, zp_encode, zs_encode, zt_encode


class Zone(Element):
    """Class representing a Zone"""

    def __init__(self, index, elk):
        super().__init__(index, elk)
        self.definition = 0
        self.area = -1
        self.bypassed = False
        self.logical_status = 0
        self.physical_status = 0
        self.voltage = 0
        self.temperature = -60
        self.triggered_alarm = False

    def __str__(self):
        return (
            "{indx:d} '{name}' type:{typ} status:{logl}/{phys}"
            " area:{area:d} trig:{trig} v:{volt} temp:{temp}"
        ).format(
            name=self.name,
            indx=self._index,
            typ=ZoneType(self.definition).name,
            logl=ZoneLogicalStatus(self.logical_status).name,
            trig=self.triggered_alarm,
            volt=self.voltage,
            area=self.area,
            temp=self.temperature,
            phys=ZonePhysicalStatus(self.physical_status).name,
        )

    def zone_trigger(self):
        """(Helper) Trigger zone."""
        self._elk.send(zt_encode(self._index))


class Zones(Elements):
    """Handling for multiple zones"""

    def __init__(self, elk):
        super().__init__(elk, Zone, Max.ZONES.value)
        elk.add_handler("AZ", self._az_handler)
        elk.add_handler("LW", self._lw_handler)
        elk.add_handler("ST", self._st_handler)
        elk.add_handler("ZB", self._zb_handler)
        elk.add_handler("ZC", self._zc_handler)
        elk.add_handler("ZD", self._zd_handler)
        elk.add_handler("ZP", self._zp_handler)
        elk.add_handler("ZS", self._zs_handler)
        elk.add_handler("ZV", self._zv_handler)

    def sync(self):
        """Retrieve zones from ElkM1"""
        self.elk.send(az_encode())
        self.elk.send(zd_encode())
        self.elk.send(zp_encode())
        self.elk.send(zs_encode())
        self.get_descriptions(TextDescriptions.ZONE.value)

    def _az_handler(self, alarm_status):
        for zone in self.elements:
            zone.setattr("triggered_alarm", alarm_status[zone.index] != "0", True)

    # pylint: disable=unused-argument
    def _lw_handler(self, keypad_temps, zone_temps):
        for i in range(16):
            zone = self.elements[i]
            if zone_temps[zone.index] > -60:
                zone.setattr("temperature", zone_temps[zone.index], True)

    def _st_handler(self, group, device, temperature):
        if group == 0:
            self.elements[device].setattr("temperature", temperature, True)

    def _zb_handler(self, zone_number, zone_bypassed):
        self.elements[zone_number].setattr("bypassed", zone_bypassed, True)

    def _zc_handler(self, zone_number, zone_status):
        self.elements[zone_number].setattr("logical_status", zone_status[0], False)
        self.elements[zone_number].setattr("physical_status", zone_status[1], True)

    def _zd_handler(self, zone_definitions):
        for zone in self.elements:
            zone.setattr("definition", zone_definitions[zone.index], True)

    def _zp_handler(self, zone_partitions):
        for zone in self.elements:
            zone.setattr("area", zone_partitions[zone.index], True)

    def _zs_handler(self, zone_statuses):
        for zone in self.elements:
            zone.setattr("logical_status", zone_statuses[zone.index][0], False)
            zone.setattr("physical_status", zone_statuses[zone.index][1], True)

    def _zv_handler(self, zone_number, zone_voltage):
        self.elements[zone_number].setattr("voltage", zone_voltage, True)



"""Definition of an ElkM1 Area"""
from .const import ElkRPStatus
from .elements import Element
from .message import vn_encode, lw_encode, sw_encode, sp_encode, ss_encode


class Panel(Element):
    """Class representing an Area"""

    def __init__(self, elk):
        super().__init__(0, elk)
        self.real_time_clock = None
        self.elkm1_version = None
        self.xep_version = None
        self.remote_programming_status = 0
        self.system_trouble_status = ""
        self.setattr("name", "ElkM1", True)
        self._elk.add_sync_handler(self.sync)

    def sync(self):
        """Retrieve panel information from ElkM1"""
        self._elk.add_handler("VN", self._vn_handler)
        self._elk.add_handler("XK", self._xk_handler)
        self._elk.add_handler("RP", self._rp_handler)
        self._elk.add_handler("IE", self._elk.call_sync_handlers)
        self._elk.add_handler("SS", self._ss_handler)
        self._elk.send(vn_encode())
        self._elk.send(lw_encode())
        # We always send ss last so we know when
        # the sync is completed
        self._elk.send(ss_encode())

    def speak_word(self, word):
        """(Helper) Speak word."""
        self._elk.send(sw_encode(word))

    def speak_phrase(self, phrase):
        """(Helper) Speak phrase."""
        self._elk.send(sp_encode(phrase))

    def _vn_handler(self, elkm1_version, xep_version):
        self.setattr("elkm1_version", elkm1_version, False)
        self.setattr("xep_version", xep_version, True)

    def _xk_handler(self, real_time_clock):
        self.setattr("real_time_clock", real_time_clock, True)

    def _ss_handler(self, system_trouble_status):
        def _get_status(index, trouble, zone_encoded=False):
            if system_trouble_status[index] != "0":
                if zone_encoded:
                    zone = ord(system_trouble_status[index]) - 0x30
                    statuses.append("{} zone {}".format(trouble, zone))
                else:
                    statuses.append(trouble)

        statuses = []
        _get_status(0, "AC Fail")
        _get_status(1, "Box Tamper", True)
        _get_status(2, "Fail To Communicate")
        _get_status(3, "EEProm Memory Error")
        _get_status(4, "Low Battery Control")
        _get_status(5, "Transmitter Low Battery", True)
        _get_status(6, "Over Current")
        _get_status(7, "Telephone Fault")
        _get_status(9, "Output 2")
        _get_status(10, "Missing Keypad")
        _get_status(11, "Zone Expander")
        _get_status(12, "Output Expander")
        _get_status(14, "ELKRP Remote Access")
        _get_status(16, "Common Area Not Armed")
        _get_status(17, "Flash Memory Error")
        _get_status(18, "Security Alert", True)
        _get_status(19, "Serial Port Expander")
        _get_status(20, "Lost Transmitter", True)
        _get_status(21, "GE Smoke CleanMe")
        _get_status(22, "Ethernet")
        _get_status(31, "Display Message In Keypad Line 1")
        _get_status(32, "Display Message In Keypad Line 2")
        _get_status(33, "Fire", True)
        self.setattr("system_trouble_status", ", ".join(statuses), True)

    def _rp_handler(self, remote_programming_status):
        if remote_programming_status == ElkRPStatus.DISCONNECTED.value:
            self._elk.resume()
        else:
            self._elk.pause()
        self.setattr("remote_programming_status", remote_programming_status, True)



"""Utility functions"""

import logging
import ssl

LOG = logging.getLogger(__name__)


def url_scheme_is_secure(url):
    """Check if the URL is one that requires SSL/TLS."""
    scheme, _dest = url.split("://")
    return scheme == "elks"


def parse_url(url):
    """Parse a Elk connection string """
    scheme, dest = url.split("://")
    host = None
    ssl_context = None
    if scheme == "elk":
        host, port = dest.split(":") if ":" in dest else (dest, 2101)
    elif scheme == "elks":
        host, port = dest.split(":") if ":" in dest else (dest, 2601)
        ssl_context = ssl.SSLContext(ssl.PROTOCOL_TLSv1)
        ssl_context.verify_mode = ssl.CERT_NONE
    elif scheme == "serial":
        host, port = dest.split(":") if ":" in dest else (dest, 115200)
    else:
        raise ValueError("Invalid scheme '%s'" % scheme)
    return (scheme, host, int(port), ssl_context)


def pretty_const(value):
    """Make a constant pretty for printing in GUI"""
    words = value.split("_")
    pretty = words[0].capitalize()
    for word in words[1:]:
        pretty += " " + word.lower()
    return pretty


def username(elk, user_number):
    """Return name of user."""
    if user_number >= 0 and user_number < elk.users.max_elements:
        return elk.users[user_number].name
    if user_number == 201:
        return "*Program*"
    if user_number == 202:
        return "*Elk RP*"
    if user_number == 203:
        return "*Quick arm*"
    return ""



"""
  Base of all the elements found on the Elk panel... Zone, Keypad, etc.
"""

from abc import abstractmethod
from .message import sd_encode


class Element:
    """Element class"""

    def __init__(self, index, elk):
        self._index = index
        self._elk = elk
        self._callbacks = []
        self.name = self.default_name()
        self._changeset = {}
        self._configured = False

    @property
    def index(self):
        """Get the index, immutable once class created"""
        return self._index

    @property
    def configured(self):
        """If a callback has ever been triggered this will be true."""
        return self._configured

    def add_callback(self, callback):
        """Callbacks when attribute of element changes"""
        self._callbacks.append(callback)

    def remove_callback(self, callback):
        """Callbacks when attribute of element changes"""
        if callback in self._callbacks:
            self._callbacks.remove(callback)

    def _call_callbacks(self):
        """Callbacks when attribute of element changes"""
        self._configured = True
        for callback in self._callbacks:
            callback(self, self._changeset)
        self._changeset = {}

    def setattr(self, attr, new_value, close_the_changeset=True):
        """If attribute value has changed then set it and call the callbacks"""
        existing_value = getattr(self, attr, None)
        if existing_value != new_value:
            setattr(self, attr, new_value)
            self._changeset[attr] = new_value

        if close_the_changeset and self._changeset:
            self._call_callbacks()

    def default_name(self, separator="-"):
        """Return a default name for based on class and index of element"""
        return self.__class__.__name__ + "{}{:03d}".format(separator, self._index + 1)

    def is_default_name(self):
        """Check if the name assigned is the default_name"""
        return self.name == self.default_name()

    def __str__(self):
        varlist = {
            k: v
            for (k, v) in vars(self).items()
            if not k.startswith("_") and k != "name"
        }.items()
        varstr = " ".join("%s:%s" % item for item in varlist)
        return "{} '{}' {}".format(self._index, self.name, varstr)

    def as_dict(self):
        """Package up the public attributes as a dict."""
        attrs = vars(self)
        return {key: attrs[key] for key in attrs if not key.startswith("_")}


class Elements:
    """Base for list of elements."""

    def __init__(self, elk, class_, max_elements):
        self.elk = elk
        self.max_elements = max_elements
        self.elements = [class_(i, elk) for i in range(max_elements)]
        self.elk.add_sync_handler(self.sync)

    def __iter__(self):
        for element in self.elements:
            yield element

    def __getitem__(self, key):
        return self.elements[key]

    def _got_desc(self, descriptions):
        for element in self.elements:
            if element.index >= len(descriptions):
                break
            if descriptions[element.index] is not None:
                element.setattr("name", descriptions[element.index], True)

    def get_descriptions(self, description_type):
        """
        Gets the descriptions for specified type.
        When complete the callback is called with a list of descriptions
        """
        (desc_type, max_units) = description_type
        results = [None] * max_units
        self.elk._descriptions_in_progress[desc_type] = (
            max_units,
            results,
            self._got_desc,
        )
        self.elk.send(sd_encode(desc_type=desc_type, unit=0))

    @abstractmethod
    def sync(self):
        """Synchronize elements"""
        pass



"""Definition of an ElkM1 Keypad."""
import datetime as dt

from .const import Max, TextDescriptions
from .elements import Element, Elements
from .message import ka_encode


class Keypad(Element):
    """Class representing an Keypad"""

    def __init__(self, index, elk):
        super().__init__(index, elk)
        self.area = -1
        self.temperature = -40
        self.last_user_time = dt.datetime.now(dt.timezone.utc)
        self.last_user = -1
        self.code = ""
        self.last_keypress = None


class Keypads(Elements):
    """Handling for multiple areas"""

    def __init__(self, elk):
        super().__init__(elk, Keypad, Max.KEYPADS.value)
        elk.add_handler("IC", self._ic_handler)
        elk.add_handler("KA", self._ka_handler)
        elk.add_handler("KC", self._kc_handler)
        elk.add_handler("LW", self._lw_handler)
        elk.add_handler("ST", self._st_handler)

    def sync(self):
        """Retrieve areas from ElkM1"""
        self.elk.send(ka_encode())
        self.get_descriptions(TextDescriptions.KEYPAD.value)

    # pylint: disable=unused-argument
    def _ic_handler(self, code, user, keypad):
        keypad_ = self.elements[keypad]

        # By setting a time this will force the IC change to always be reported
        keypad_.setattr("last_user_time", dt.datetime.now(dt.timezone.utc), False)

        # If user is negative then invalid code entered
        keypad_.setattr("code", code if user < 0 else "****", False)
        keypad_.setattr("last_user", user, True)

    def _ka_handler(self, keypad_areas):
        for keypad in self.elements:
            if keypad_areas[keypad.index] >= 0:
                keypad.setattr("area", keypad_areas[keypad.index], True)

    def _kc_handler(self, keypad, key):
        if key > 0:
            self.elements[keypad].setattr("last_keypress", key, True)

    # pylint: disable=unused-argument
    def _lw_handler(self, keypad_temps, zone_temps):
        for keypad in self.elements:
            if keypad_temps[keypad.index] > -40:
                keypad.setattr("temperature", keypad_temps[keypad.index], True)

    def _st_handler(self, group, device, temperature):
        if group == 1:
            self.elements[device].setattr("temperature", temperature, True)



"""Definition of an ElkM1 Area"""

from .const import Max, TextDescriptions
from .elements import Element, Elements
from .message import as_encode, az_encode, al_encode, dm_encode


class Area(Element):
    """Class representing an Area"""

    def __init__(self, index, elk):
        super().__init__(index, elk)
        self.armed_status = None
        self.arm_up_state = None
        self.alarm_state = None
        self.alarm_memory = None
        self.is_exit = False
        self.timer1 = 0
        self.timer2 = 0
        self.last_log = None

    def arm(self, level, code):
        """(Helper) Arm system at specified level (away, vacation, etc)"""
        self._elk.send(al_encode(level, self._index, code))

    def disarm(self, code):
        """(Helper) Disarm system."""
        self.arm(0, code)

    def display_message(self, clear, beep, timeout, line1, line2):
        """Display a message on all of the keypads in this area."""
        self._elk.send(dm_encode(self._index, clear, beep, timeout, line1, line2))


class Areas(Elements):
    """Handling for multiple areas"""

    def __init__(self, elk):
        super().__init__(elk, Area, Max.AREAS.value)
        elk.add_handler("AM", self._am_handler)
        elk.add_handler("AS", self._as_handler)
        elk.add_handler("EE", self._ee_handler)
        elk.add_handler("LD", self._ld_handler)

    def sync(self):
        """Retrieve areas from ElkM1"""
        self.elk.send(as_encode())
        self.get_descriptions(TextDescriptions.AREA.value)

    def _am_handler(self, alarm_memory):
        for area in self.elements:
            area.setattr("alarm_memory", alarm_memory[area.index], True)

    def _as_handler(self, armed_statuses, arm_up_states, alarm_states):
        update_alarm_triggers = False
        for area in self.elements:
            area.setattr("armed_status", armed_statuses[area.index], False)
            area.setattr("arm_up_state", arm_up_states[area.index], False)
            if (
                area.alarm_state != alarm_states[area.index]
                or alarm_states[area.index] != "0"
            ):
                update_alarm_triggers = True
            area.setattr("alarm_state", alarm_states[area.index], True)

        if update_alarm_triggers:
            self.elk.send(az_encode())

    # pylint: disable=too-many-arguments
    def _ee_handler(self, area, is_exit, timer1, timer2, armed_status):
        area = self.elements[area]
        area.setattr("armed_status", armed_status, False)
        area.setattr("timer1", timer1, False)
        area.setattr("timer2", timer2, False)
        area.setattr("is_exit", is_exit, True)

    # ElkM1 global setting G35 must be set for LD messages to be sent
    def _ld_handler(self, area, log):
        if log["event"] in [1173, 1174]:
            # arm/disarm log (YAGNI - decode number for more log types when needed)
            log["user_number"] = log["number"]
        self.elements[area].setattr("last_log", log, True)



"""Master class that combines all ElkM1 pieces together."""

import asyncio
from functools import partial
import logging
from importlib import import_module
import serial_asyncio

from .message import MessageDecode, sd_encode, ua_encode
from .proto import Connection
from .util import parse_url, url_scheme_is_secure

LOG = logging.getLogger(__name__)


class Elk:
    """Represents all the components on an Elk panel."""

    def __init__(self, config, loop=None):
        """Initialize a new Elk instance."""
        self.loop = loop if loop else asyncio.get_event_loop()
        self._config = config
        self._conn = None
        self._transport = None
        self.connection_lost_callbk = None
        self._connection_retry_timer = 1
        self._message_decode = MessageDecode()
        self._sync_handlers = []
        self._descriptions_in_progress = {}
        self._sync_event = asyncio.Event()
        self._heartbeat = None
        self._invalid_auth = False
        self._connection_retry_task = None
        self._disconnect_requested = False

        self.add_handler("XK", self._xk_handler)
        self.add_handler("SD", self._sd_handler)

        # Setup for all the types of elements tracked
        if "element_list" in config:
            self.element_list = config["element_list"]
        else:
            self.element_list = [
                "panel",
                "zones",
                "lights",
                "areas",
                "tasks",
                "keypads",
                "outputs",
                "thermostats",
                "counters",
                "settings",
                "users",
            ]
        # We send a "ua" command at the end of the
        # sync to indicate the sync is done.
        self.add_handler("UA", self._sync_complete)

        for element in self.element_list:
            self._create_element(element)

    def _sync_complete(self, **kwargs):
        self._sync_event.set()

    def _create_element(self, element):
        module = import_module("elkm1_lib." + element)
        class_ = getattr(module, element.capitalize())
        setattr(self, element, class_(self))

    async def _connect(self, connection_lost_callbk=None):
        """Asyncio connection to Elk."""
        self.connection_lost_callbk = connection_lost_callbk
        self._invalid_auth = False
        url = self._config["url"]
        LOG.info("Connecting to ElkM1 at %s", url)
        scheme, dest, param, ssl_context = parse_url(url)
        conn = partial(
            Connection,
            self.loop,
            self._connected,
            self._disconnected,
            self._got_data,
            self._timeout,
        )
        try:
            if scheme == "serial":
                await serial_asyncio.create_serial_connection(
                    self.loop, conn, dest, baudrate=param
                )
            else:
                await asyncio.wait_for(
                    self.loop.create_connection(
                        conn, host=dest, port=param, ssl=ssl_context
                    ),
                    timeout=30,
                )
        except (ValueError, OSError, asyncio.TimeoutError) as err:
            if self._disconnect_requested:
                # disconnect called while trying to reconnect
                return
            LOG.warning(
                "Could not connect to ElkM1 (%s). Retrying in %d seconds",
                err,
                self._connection_retry_timer,
            )
            self._connection_retry_task = self.loop.call_later(
                self._connection_retry_timer, self.connect
            )
            self._connection_retry_timer = (
                2 * self._connection_retry_timer
                if self._connection_retry_timer < 32
                else 60
            )

    def _connected(self, transport, conn):
        """Login and sync the ElkM1 panel to memory."""
        LOG.info("Connected to ElkM1")
        self._conn = conn
        self._transport = transport
        self._connection_retry_timer = 1
        if url_scheme_is_secure(self._config["url"]):
            self._conn.write_data(self._config["userid"], raw=True)
            self._conn.write_data(self._config["password"], raw=True)
        self.call_sync_handlers()
        if not self._config["url"].startswith("serial://"):
            self._heartbeat = self.loop.call_later(120, self._reset_connection)

    def _reset_connection(self):
        LOG.warning("ElkM1 connection heartbeat timed out, disconnecting")
        self._transport.close()
        self._heartbeat = None

    # pylint: disable=unused-argument
    def _xk_handler(self, real_time_clock):
        if not self._heartbeat:
            return
        self._heartbeat.cancel()
        self._heartbeat = self.loop.call_later(120, self._reset_connection)

    def _disconnected(self):
        LOG.warning("ElkM1 at %s disconnected", self._config["url"])
        self._conn = None
        self._connection_retry_task = self.loop.call_later(
            self._connection_retry_timer, self.connect
        )
        if self._heartbeat:
            self._heartbeat.cancel()
            self._heartbeat = None

    def add_handler(self, msg_type, handler):
        self._message_decode.add_handler(msg_type, handler)

    def _got_data(self, data):  # pylint: disable=no-self-use
        LOG.debug("got_data '%s'", data)
        try:
            self._message_decode.decode(data)
        except (ValueError, AttributeError) as err:
            if (
                not len(data)
                or data.startswith("Username: ")
                or data.startswith("Password: ")
            ):
                return
            elif data.startswith("Username/Password not found"):
                LOG.error("Invalid username or password.")
                self.disconnect()
                self._invalid_auth = True
                self._sync_event.set()
            elif "Login successful" in data:
                LOG.info("Successful login.")
                return

            LOG.debug(err)

    def _timeout(self, msg_code):
        self._message_decode.timeout_handler(msg_code)

    def add_sync_handler(self, sync_handler):
        """Register a fn that synchronizes part of the panel."""
        self._sync_handlers.append(sync_handler)

    def call_sync_handlers(self):
        """Invoke the synchronization handlers."""
        LOG.debug("Synchronizing panel...")
        self._sync_event.clear()
        for sync_handler in self._sync_handlers:
            sync_handler()
        # We send an empty "ua" message at the end of the sync
        # so we can watch for the "UA" response to come back
        # and know that the sync has completed.
        self.send(ua_encode(0))

    async def sync_complete(self):
        return await self._sync_event.wait()

    # pylint: disable=unused-argument
    def _sd_handler(self, desc_type, unit, desc, show_on_keypad):
        """Text description"""
        if desc_type not in self._descriptions_in_progress:
            LOG.debug("Text description response ignored for " + str(desc_type))
            return

        (max_units, results, callback) = self._descriptions_in_progress[desc_type]
        if unit < 0 or unit >= max_units:
            callback(results)
            del self._descriptions_in_progress[desc_type]
            return

        results[unit] = desc
        self.send(sd_encode(desc_type=desc_type, unit=unit + 1))

    def is_connected(self):
        """Status of connection to Elk."""
        return self._conn is not None

    def connect(self):
        """Connect to the panel"""
        self._disconnect_requested = False
        asyncio.ensure_future(self._connect())

    def run(self):
        """Enter the asyncio loop."""
        self.loop.run_forever()

    def send(self, msg):
        """Send a message to Elk panel."""
        if self._conn:
            self._conn.write_data(msg.message, msg.response_command)

    def pause(self):
        """Pause the connection from sending/receiving."""
        if self._conn:
            self._conn.pause()

    def resume(self):
        """Restart the connection from sending/receiving."""
        if self._conn:
            self._conn.resume()

    def disconnect(self):
        """Disconnect the connection from sending/receiving."""
        self._disconnect_requested = True
        if self._conn:
            self._conn.stop()
            self._conn = None
        if self._connection_retry_task:
            self._connection_retry_task.cancel()
            self._connection_retry_task = None
        if self._transport:
            self._transport.close()
            self._transport = None
        if self._heartbeat:
            self._heartbeat.cancel()
            self._heartbeat = None

    @property
    def invalid_auth(self):
        """Last session was disconnected due to invalid auth details."""
        return self._invalid_auth
